/*
 *  $Id: adi2probe.c,v 1.6 2005/06/03 01:43:46 loic Exp $
 *
 *  (C) 1995 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include <stdio.h>

#include "mxmpi.h"
#include "mpid.h"
#include "mpiddev.h"
#include "mpid_debug.h"
#include "../util/queue.h"

void MPID_Iprobe( 
	struct MPIR_COMMUNICATOR *comm_ptr, 
	int tag, 
	int context_id, 
	int src_lrank, 
	int *found, 
	int *error_code, 
	MPI_Status *status )
{
  mx_return_t rc;
  uint64_t match_mask;
  uint64_t match_val;
  mx_status_t mx_status;
  uint32_t result;

  /* Figure out match value and mask */
  if (tag == MPI_ANY_TAG) {
    if (src_lrank == MPI_ANY_SOURCE) {
      match_mask = MXMPI_MATCH_MASK_SRC_TAG;
      match_val = MXMPI_MATCH(context_id, 0, 0);
    } else {
      match_mask = MXMPI_MATCH_MASK_TAG;
      match_val = MXMPI_MATCH(context_id, src_lrank, 0);
    }
  } else {
    if (src_lrank == MPI_ANY_SOURCE) {
      match_mask = MXMPI_MATCH_MASK_SRC;
      match_val = MXMPI_MATCH(context_id, 0, tag);
    } else {
      match_mask = MX_MATCH_MASK_NONE;
      match_val = MXMPI_MATCH(context_id, src_lrank, tag);
    }
  }

  /* Probe for the message */
  rc = mx_iprobe(mxmpi.my_endpoint, 
  	match_val, match_mask, &mx_status, &result);

  MXMPI_NOTE(("iprobe match=%x:%x mask=%d, result=%s\n",
			context_id, match_val, match_mask,
			(rc == MX_SUCCESS) ? "yes" :
			(rc == MX_NO_MATCH) ? "no" : "error"));

  if (rc != MX_SUCCESS) {
    fprintf(stderr, "[%d] iprobe failed: %s\n", 
	MPID_MyWorldRank,
	mx_strerror(rc));
    mxmpi_abort(0);
  }

  /* successful call, result tells whether we found anything */
  if (result) {
    *found = 1;
    status->count = mx_status.msg_length;
    status->MPI_TAG = MXMPI_EXTRACT_TAG(mx_status.match_info);
    status->MPI_SOURCE = MXMPI_EXTRACT_SRC(mx_status.match_info);

  } else {
    *found = 0;
  }
}

void MPID_Probe( 
	struct MPIR_COMMUNICATOR *comm_ptr, 
	int tag, 
	int context_id, 
	int src_lrank, 
	int *error_code, 
	MPI_Status *status )
{
    int found;

    *error_code = 0;
    DEBUG_PRINT_MSG("Entering Probe");
    DEBUG_PRINT_MSG( "Entering while(1)" );
    while (1) {
	MPID_Iprobe( comm_ptr, tag, context_id, src_lrank, &found, error_code, 
		     status );
	if (found || *error_code) break;
	/* Wait for a message */
	MPID_DeviceCheck( MPID_BLOCKING );
    }
    DEBUG_PRINT_MSG( "Leaving while(1)" );
    DEBUG_PRINT_MSG("Exiting Probe");
}
