
static void
gmpi_tcpinit(char *buffer, size_t buflen, char *master_name, int master_port,
	     struct sockaddr_in *master_addr,
	     unsigned u1, unsigned u2, unsigned u3, unsigned u4, unsigned u5,
	     unsigned u6, unsigned u7)
{
  int gmpi_sockfd, gmpi_sockfd2;
  socklen_t addrlen;
  struct sockaddr_in slave_addr;
  int slave_port;
  struct hostent *master;
  struct timeval stop_time, start_time;
  size_t count;

  /* get a socket */
  gmpi_sockfd = socket (AF_INET, SOCK_STREAM, 0);
  if (gmpi_sockfd < 0) {
    fprintf (stderr, "[%u] Error: Unable to open a socket !\n", 
	     MPID_MyWorldRank);
    gmpi_abort (0);
  }
  
  /* bind the socket to a port */
  memset (&slave_addr, 0, sizeof(slave_addr));
  slave_addr.sin_family = AF_INET;
  slave_addr.sin_addr.s_addr = INADDR_ANY;  
  slave_addr.sin_port = 0;
  addrlen = sizeof(slave_addr);
  if ((bind(gmpi_sockfd, (struct sockaddr *) &(slave_addr), 
	    sizeof (slave_addr)) != 0)
      || (getsockname(gmpi_sockfd,(struct sockaddr *) &(slave_addr),
		      &addrlen) != 0)
      || slave_addr.sin_port == 0) {
    fprintf (stderr, "[%u] Error: Unable to bind to a socket:%s\n",
	     MPID_MyWorldRank, strerror(errno));
    gmpi_abort (0);
  }
  slave_port = ntohs(slave_addr.sin_port);
  
  /* listen for the master */
  if (listen(gmpi_sockfd, 1) != 0) {
    fprintf (stderr, "[%u] Error: Error when listening on the socket !\n", 
	     MPID_MyWorldRank);
    gmpi_abort (0);
  }
  
  /* get another socket */
  gmpi_sockfd2 = socket (AF_INET, SOCK_STREAM, 0);
  if (gmpi_sockfd2 < 0) {
    fprintf (stderr, "[%u] Error: Unable to open a socket (2) !\n", 
	     MPID_MyWorldRank);
    gmpi_abort (0);
  }
  
  /* get the master's IP address */
  master = gethostbyname (master_name);
  if (master == NULL) { 
    fprintf (stderr, "[%u] Error: Unable to translate "
	     "the hostname of the master (%s)!\n", 
	     MPID_MyWorldRank, master_name);
    gmpi_abort (0);
  }
  
  /* connect to the master */
  memset(master_addr,0, sizeof (*master_addr));
  master_addr->sin_family = AF_INET;
  memcpy(&master_addr->sin_addr, master->h_addr,
	 master->h_length);
  master_addr->sin_port = htons(master_port);
  gettimeofday(&start_time, NULL);
  while (connect (gmpi_sockfd2, (struct sockaddr *)master_addr, 
		  sizeof (*master_addr)) < 0) {
	gettimeofday(&stop_time, NULL);
	if ((stop_time.tv_sec - start_time.tv_sec) > GMPI_INIT_TIMEOUT) {
	  fprintf (stderr, "[%u] Error: Unable to connect to "
		   "the master:%s !\n", MPID_MyWorldRank, strerror(errno));
	  gmpi_abort (0);
	}
	usleep(random()%10000);
  }
  
  /*
   * Send the message to the master:
   * <<<magic:ID:port/board:mac_high:mac_low:numanode:pid::socket_port>>>
   */
  count = 0;
  snprintf(buffer, buflen, 
	   "<<<%u:%u:%u:%u:%u:%u:%u::%u>>>\n",
	   u1, u2, u3, u4, u5, u6, u7, slave_port);
  buffer[buflen - 1] = 0;
  
	   /*  magic_number, 
	   MPID_MyWorldRank, (port_id + (board_id << 16)), 
	   (gmpi.my_unique_id[1] + (gmpi.my_unique_id[0] << 8)),
	   (gmpi.my_unique_id[5] + (gmpi.my_unique_id[4] << 8)
	    + (gmpi.my_unique_id[3] << 16) + (gmpi.my_unique_id[2] << 24)),
	   gmpi.numa_nodes[MPID_MyWorldRank], 
	   (int) getpid (), slave_port);
	   */
  while (count < strlen (buffer)) {
    int k;
    k = write (gmpi_sockfd2, &(buffer[count]), strlen (buffer) - count);
    if (k <= 0) {
      fprintf (stderr, "[%u] Error: write to socket failed,rc=%d:%s\n", 
	       MPID_MyWorldRank, k ,strerror(errno));
      gmpi_abort (0);
    }
    count += k;
  }
  close (gmpi_sockfd2);
  
  /* wait for the map from the master */
  gmpi_sockfd2 = accept (gmpi_sockfd, 0, 0);
  if (gmpi_sockfd2 < 0) {
    fprintf (stderr, "[%u] Error: Error on socket accept !\n", 
	     MPID_MyWorldRank);
    gmpi_abort (0);
  }
  
  /* Get the whole GM mapping from the master */
  count = 0;
  memset (buffer, -1, buflen);
  buflen -= 1;
  buffer[buflen] = 0;
  while (count < 3 || memcmp (buffer+count -3, "]]]", 3) != 0) {
    int k;
    if (count >= buflen) {
      fprintf(stderr,"[%u] Error: information buffer too big: too many nodes?\n",
	      MPID_MyWorldRank);
      gmpi_abort(0);
    }
    k = read (gmpi_sockfd2, &(buffer[count]), 
	      buflen - count);
    if (k <= 0) {
      fprintf (stderr, "[%u] Error: read from socket failed,rc=%d:%s !\n", 
	       MPID_MyWorldRank, k, strerror(errno));
      gmpi_abort (0);
    }
    count += k;
  }
  close (gmpi_sockfd2);
}
