/*
 *  $Id: mxmpi_cancel.c,v 1.3 2007/01/31 19:02:42 loic Exp $
 *
 *  (C) 1996 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include <stdlib.h>

#include "mxmpi.h"
#include "mpid.h"

/*
 * This file contains the routines to handle canceling a message
 *
 * Note: for now, cancel will probably only work on unmatched receives.
 * However, this code provides the hooks for supporting more complete
 * cancel implementations.
 */

void MPID_SendCancel( MPI_Request request, int *error_code )
{
  MXMPI_NOTIMPL();
#if 0
	/* begin MPID_SendCancel */

    MPIR_SHANDLE *shandle = &request->shandle;

    DEBUG_PRINT_MSG("S Starting SendCancel");

    shandle->is_cancelled = 0;
    shandle->cancel_complete = 0;

    MPID_SendCancelPacket(&request, error_code); 

    if (*error_code == MPI_SUCCESS) {   /* begin if !fail */
	DEBUG_PRINT_MSG("Entering while !shandle->cancel_complete");
	while (!shandle->cancel_complete) {  /* begin while */
	    MPID_DeviceCheck( MPID_BLOCKING );
	}
	DEBUG_PRINT_MSG("Leaving while !shandle->cancel_complete");

	if (shandle->is_cancelled) { 
	    if (shandle->finish)
		(shandle->finish)(shandle);  
	    if (shandle->handle_type == MPIR_PERSISTENT_SEND) {
		MPIR_PSHANDLE *pshandle = (MPIR_PSHANDLE *)request;
		pshandle->active = 0; 
	    }
	}

    }  /* end if !fail */

    DEBUG_PRINT_MSG("E Exiting SendCancel");
#endif
}  /* end MPID_SendCancel */

 
void MPID_RecvCancel( MPI_Request request, int *error_code )
{
#if !defined MX_API || MX_API < 0x203
  MXMPI_NOTIMPL();
#else
  MPI_Status status;
  mx_return_t rc;
  int32_t flag;
  MPIR_RHANDLE * rhandle = &request->rhandle;

  rc = mx_cancel(mxmpi.my_endpoint, &rhandle->mx_req_handle, &flag);
  if (rc != MX_SUCCESS) {
    fprintf(stderr, "[%d]: error from mx_cancel in RecvCancel: %s\n", 
	    MPID_MyWorldRank, mx_strerror(rc));
    mxmpi_abort(0);
  }
  
  if (flag) {
    rhandle->is_complete = 1;
    /* Call any finishing routine needed */
    if (rhandle->finish)  {
      (rhandle->finish)( rhandle );
    }
    rhandle->mx_req_handle = 0;
    
    /* Fill in rhandle fields from status */
    rhandle->s.count = 0;
    rhandle->s.MPI_TAG = MPIR_MSG_CANCELLED;
    rhandle->s.MPI_ERROR = MPI_SUCCESS;
  }
  *error_code = MPI_SUCCESS;
#endif
}
