/*************************************************************************
 * Myricom MPICH-MX ch_mx backend                                        *
 * Copyright (c) 2003 by Myricom, Inc.                                   *
 * All rights reserved.                                                  *
 *************************************************************************/

#include <sys/utsname.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <unistd.h>

#include "mxmpi.h"

#include "mpid.h"
#include "mpiddev.h"
#include "mpimem.h"
#include "reqalloc.h"
#include "../util/cmnargs.h"
#include "../util/queue.h"

#include "sendq.h"    /* For MPIR_FORGET_SEND */

#include "mxmpi_smpi.h"
#include "mxmpi_inline.h"

/* Home for these globals */
int MPID_MyWorldSize, MPID_MyWorldRank;
int MPID_Print_queues = 0;
MPID_SBHeader MPIR_rhandles;
MPID_SBHeader MPIR_shandles;

/* MX specific globals */
struct mxmpi_var mxmpi;
struct smpi_var smpi;

#if defined(USE_HOLD_LAST_DEBUG) || defined(MPID_DEBUG_ALL)
char ch_debug_buf[CH_MAX_DEBUG_LINE];
#endif

#if 0

/***************************************************************************/
/* Some operations are completed in several stages.  To ensure that a      */
/* process does not exit from MPID_End while requests are pending, we keep */
/* track of how many are out-standing                                      */
/***************************************************************************/
int MPID_n_pending = 0;

/* This is a prototype for this function used to get the configuration */
MPID_Config * MPID_GetConfigInfo (int *, char ***, void * ,int *);
#endif

mx_endpoint_t mxmpi_endpoint;
mx_endpoint_addr_t mxmpi_ep_addr;

void MPID_Init(int *argc, char ***argv, void *config, int *error_code)
{
  mx_return_t rc;

  rc = mx_init();
  if (rc != MX_SUCCESS) {
    fprintf(stderr, "mx_initialize failed, rc = %d\n", rc);
    *error_code = MPI_ERR_INTERN;
    return;
  }

  (void) MPID_GetConfigInfo(argc, argv, config, error_code);

    /* Initialize the send/receive handle allocation system */
    /* Use the persistent version of send/receive (since we don't have
       separate MPIR_pshandles/MPIR_prhandles) */
    MPIR_shandles   = MPID_SBinit( sizeof( MPIR_PSHANDLE ), 100, 100 );
    MPIR_rhandles   = MPID_SBinit( sizeof( MPIR_PRHANDLE ), 100, 100 );
    
  /* All successful */
  *error_code = MPI_SUCCESS;
  return;
#if 0
    int i, j, np, err, msgrp = 0, mypid;
    MPID_Device *dev;
    char hostname[SYS_NMLN];
    char execname[256];
    struct MPIR_COMMUNICATOR *comm = 0;
    MPI_Status status;    
    MPID_Config *config_info;

    /* here we bypass the default fake config function. Instead, we call 
       the MPID_GetConfigInfo to fill up the MPID_Config structure to describe
       the different devices and the routes. Seems that MPICH dudes didn't
       want to use it, but everything is there. Strange. */
    config_info = MPID_GetConfigInfo(argc, argv, config, error_code);

    config_info_copy = config_info;
    /*
     * Create the device set structure.  Currently, only handles one
     * device and maps all operations to that device
     */
    MPID_devset = (MPID_DevSet *)MALLOC( sizeof(MPID_DevSet) );
    if (!MPID_devset) {
	*error_code = MPI_ERR_INTERN;
	return ;
    }
    /* Make devset safe for initializations errors */
    MPID_devset->dev_list = 0;
    
    np = 0;
    if (config_info) {
	int ndev = 0;
	MPID_Config *p;
	p = config_info;
	while (p) {
	    ndev++;
	    for (i=0; i<p->num_served; i++) {
		if (p->granks_served[i] > np) 
		    np = p->granks_served[i];
	    }
	    p = p->next;
	}
	MPID_devset->ndev      = ndev;
	MPID_devset->ndev_list = ndev-1;
	np++;
	MPID_devset->dev  = (MPID_Device **)MALLOC( 
	    np * sizeof(MPID_Device *) );
	if (!MPID_devset->dev) {
	    *error_code = MPI_ERR_INTERN;
	    return;
	}
	p = config_info;
	while (p) {
	    dev = (p->device_init)( argc, argv, MPID_Short_len, -1 );
	    if (!dev) {
		*error_code = MPI_ERR_INTERN;
		return;
	    }
	    dev->next = MPID_devset->dev_list;
	    MPID_devset->dev_list = dev;
	    for (i=0; i<p->num_served; i++) 
		MPID_devset->dev[p->granks_served[i]] = dev;
	    p = p->next;
	}
    }
    else 
      {
	fprintf (stderr, "Configuration retrieval failed\n");
	mxmpi_abort (0);
      }
#ifdef MPIR_MEMDEBUG
    MPID_trinit( MPID_MyWorldRank );
#endif
    /* 
     * Get the basic options.  Note that this must be AFTER the initialization
     * in case the initialization routine was responsible for sending the
     * arguments to other processors.
     */
    MPID_ProcessArgs( argc, argv );

    MPID_InitQueue();
    
#ifdef NEEDS_PROCESS_GROUP_FIX
    /* Call this to force each MPI job into a separate process group
       if there is no attached terminal.  This is necessary on some
       systems to overcome runtime libraries that kill a process group
       on abnormal exit (the proces group can contain the invoking 
       shell scripts and programs otherwise) */
    MPID_FixupProcessGroup();
#endif

    /* Needs to be changed for persistent handles.  A common request form? */

    MPID_devset->req_pending = 0;
    *error_code = MPI_SUCCESS;

    /* Initialize the array on node with MPI process 0 that contains
       pid information for all processes. The index of the array is the same
       as the mpi rank of the process. We associate this with a node by
       matching it to the values in mxmpi.node_name[].
    */
    mypid = (int) getpid();
    err = gethostname(hostname, sizeof(hostname));
    if (err == -1)
      {
	fprintf (stderr, "Hostname retrieval failed\n");
	mxmpi_abort (0);
      }
    strcpy (execname, ((const char **) (*argv))[0]);

    if (MPID_MyWorldRank == 0)
      {
	mxmpi.mpi_pids[0] = mypid;
	mxmpi.host_names[0] = (char *) malloc (strlen (hostname) + 1);
	mxmpi_malloc_assert(mxmpi.host_names[0],
			   "MPID_Init",
			   "malloc: host names");
	strcpy (mxmpi.host_names[0], hostname);
	mxmpi.exec_names[0] = (char *) malloc (strlen (execname) + 1);
	mxmpi_malloc_assert(mxmpi.exec_names[0],
			   "MPID_Init",
			   "malloc: exec names");
	strcpy (mxmpi.exec_names[0], execname);
	for (j=1; j<np; j++) 
	  {
	    MPID_RecvContig(comm, &(mxmpi.mpi_pids[j]), sizeof(int), j, 
			    0, 0, &status ,&err);
	    MPID_RecvContig(comm, hostname, SYS_NMLN * sizeof(char),
		            j, 0, 0, &status ,&err);
	    mxmpi.host_names[j] = (char *) malloc (strlen (hostname) + 1);
	    mxmpi_malloc_assert(mxmpi.host_names[j],
			       "MPID_Init",
			       "malloc: host names");
	    strcpy (mxmpi.host_names[j], hostname);
	    MPID_RecvContig(comm, execname, 256 * sizeof(char), 
			    j, 0, 0, &status ,&err);
	    mxmpi.exec_names[j] = (char *) malloc (strlen (execname) + 1);
	    mxmpi_malloc_assert(mxmpi.exec_names[j],
			       "MPID_Init",
			       "malloc: exec names");
	    strcpy (mxmpi.exec_names[j], execname);
	  }
      }
    else
      {
	MPID_SendContig(comm, &mypid, sizeof(int), MPID_MyWorldRank, 
			0, 0, 0, msgrp, &err);
	MPID_SendContig(comm, hostname, SYS_NMLN * sizeof(char), 
			MPID_MyWorldRank, 0, 0, 0, msgrp, &err);
	MPID_SendContig(comm, execname, 256 * sizeof(char), 
			MPID_MyWorldRank, 0, 0, 0, msgrp, &err);
	mxmpi.host_names[MPID_MyWorldRank] = (char *) malloc (strlen (hostname) + 1);
	mxmpi_malloc_assert(mxmpi.host_names[MPID_MyWorldRank],
			   "MPID_Init",
			   "malloc: host names");
	strcpy (mxmpi.host_names[MPID_MyWorldRank], hostname);
	mxmpi.exec_names[MPID_MyWorldRank] = (char *) malloc (strlen (execname) + 1);
	mxmpi_malloc_assert(mxmpi.exec_names[MPID_MyWorldRank],
			   "MPID_Init",
			   "malloc: exec names");
	strcpy (mxmpi.exec_names[MPID_MyWorldRank], execname);
      }
#endif
}


/* Barry Smith suggests that this indicate who is aborting the program.
   There should probably be a separate argument for whether it is a 
   user requested or internal abort.
 */
void MPID_Abort( 
	struct MPIR_COMMUNICATOR *comm_ptr, 
	int code, 
	char *user, 
	char *str )
{
    MPID_Device *dev;
    char abortString[256];

    fprintf( stderr, "[%d] %s Aborting program %s\n", MPID_MyWorldRank,
	     user ? user : "", str ? str : "!" );
    fflush( stderr );
    fflush( stdout );

#ifdef USE_PRINT_LAST_ON_ERROR
    MPID_Ch_dprint_last();
#endif

    /* Also flag a debugger that an abort has happened so that it can take
     * control while there's still useful state to be examined.
     * Remember, MPIR_Breakpoint is a complete no-op unless the debugger
     * is present.
     */
    sprintf(abortString, "%s Aborting program %s", user ? user : "", 
	    str ? str : "!" );
    MPIR_debug_abort_string = abortString;
    MPIR_debug_state        = MPIR_DEBUG_ABORTING;
    MPIR_Breakpoint();

    mxmpi_abort(code);
    mxmpi_abort(code);
#if 0
    /* We may be aborting before defining any devices */
    if (MPID_devset) {
	int found_dev = 0;
	dev = MPID_devset->dev_list;
	while (dev) {
	    found_dev = 1;
	    (*dev->abort)( comm_ptr, code, str );
	    dev = dev->next;
	}
	if (!found_dev) 
	    exit( code );
    }
    else {
	exit( code );
    }
#endif
}

void MPID_End()
{
  mx_close_endpoint(mxmpi.my_endpoint);
  mx_finalize();

#if MX_INTERP
  if (interp_file != NULL) {
    fprintf(interp_file, "close e1\n");
    fprintf(interp_file, "EEE\n");
    fflush(interp_file);
  }
#endif
#if 0
    MPID_Device *dev, *ndev;
    MPID_Config * config_info;

    DEBUG_PRINT_MSG("Entering MPID_End" )

    /* Finish off any pending transactions */
    /* Should this be part of the device terminate routines instead ? 
       Probably not, incase they need to be done in an arbitrary sequence */
    MPID_Complete_pending();
    
    if (MPID_GetMsgDebugFlag()) {
	MPID_PrintMsgDebug();
    }

    /* Eventually make this optional */
    
    if (MPID_Print_queues)
	MPID_Dump_queues();

/* We should really generate an error or warning message if there 
   are uncompleted operations... */
    dev = MPID_devset->dev_list;
    while (dev) {
	ndev = dev->next;
	/* Each device should free any storage it is using INCLUDING dev */
	(*dev->terminate)( dev );
	dev = ndev;
    }

    /* Clean up request handles */
    MPID_SBdestroy( MPIR_shandles );
    MPID_SBdestroy( MPIR_rhandles );

    /* Free remaining storage */
    FREE( MPID_devset->dev );
    FREE( MPID_devset );

    while (config_info_copy != NULL)
      {
	free(config_info_copy->device_init_name);
	free(config_info_copy->granks_served);
	config_info = config_info_copy;
	config_info_copy = config_info_copy->next;
	free(config_info);
      }

#if defined(MPIR_MEMDEBUG) && defined(MPID_ONLY)
    /* MPI_Finalize also does this */
    MPID_trdump( stdout );
#endif
#endif
}

/* Returns 1 if something found, -1 otherwise (if is_blocking is 
   MPID_NOTBLOCKING)  */
int MPID_DeviceCheck( 
	MPID_BLOCKING_TYPE is_blocking)
{
  mx_return_t rc;
  mx_request_t request;
  uint32_t result;
  int found;
  MPI_Request mpi_req;
  mx_status_t status;

  found = -1;

  while (1) {

    /* look for anything completed */
    rc = mx_ipeek(mxmpi.my_endpoint, &request, &result);
    if (rc != MX_SUCCESS) {
      fprintf (stderr, "[%d] Error in mx_ipeek: %s\n", MPID_MyWorldRank, rc);
      mxmpi_abort(1);
    }
    
    /* If something ready, complete it and grab the status */
    if (result) {
      found = 1;

      MXMPI_NOTE(("mx_ipeek found req=%p\n", request));

      /* complete the request */
      rc = (is_blocking ? mxmpi.test_or_wait : mx_test)(mxmpi.my_endpoint, &request, &status, &result);
      if (rc != MX_SUCCESS) {
        fprintf(stderr, "Error from test_or_wait: \"%s\"\n", mx_strerror(rc));
	mxmpi_abort(1);
      }
      if (!result) {
        fprintf(stderr, "test_or_wait after mx_ipeek failed to complete request!\n");
	mxmpi_abort(1);
      }
      if (status.context == NULL) {
        fprintf(stderr, "Orphaned request in DeviceCheck!\n");
	mxmpi_abort(1);
      }

      /* Fill in the MPI Request */
      mpi_req = (MPI_Request) status.context;

      /* If still active, fill in the fields */
      if (mpi_req->chandle.ref_count > 0) {

        if (mpi_req->handle_type == MPIR_SEND
	    || mpi_req->handle_type == MPIR_PERSISTENT_SEND) {
	  mpi_req->shandle.s.MPI_ERROR = mxmpi_status_to_mpi_code(status.code);
	  mpi_req->shandle.mx_req_handle = 0;

#if MX_INTERP
	  if (interp_file != NULL) {
	    fprintf(interp_file, "test s%x\n",
		      (unsigned int)&mpi_req->shandle.mx_req_handle);
	    fflush(interp_file);
	  }
#endif

	  /* Call any finishing routine needed */
	  if (mpi_req->shandle.finish)  {
	      (mpi_req->shandle.finish)( &mpi_req->shandle );
	  }
	} else if (mpi_req->handle_type == MPIR_RECV
	           || mpi_req->handle_type == MPIR_PERSISTENT_RECV) {
	  mpi_req->rhandle.s.count = status.xfer_length;
	  mpi_req->rhandle.s.MPI_TAG = MXMPI_EXTRACT_TAG(status.match_info);
	  mpi_req->rhandle.s.MPI_SOURCE = MXMPI_EXTRACT_SRC(status.match_info);
	  mpi_req->rhandle.s.MPI_ERROR = mxmpi_status_to_mpi_code(status.code);
	  mpi_req->rhandle.mx_req_handle = 0;

#if MX_INTERP
	  if (interp_file != NULL) {
	    fprintf(interp_file, "test r%x\n",
		      (unsigned int)&mpi_req->rhandle.mx_req_handle);
	    fflush(interp_file);
	  }
#endif

	  /* Call any finishing routine needed */
	  if (mpi_req->rhandle.finish)  {
	      (mpi_req->rhandle.finish)( &mpi_req->rhandle );
	  }
	} else {
	  fprintf(stderr, "Mysterious active handle_type = %d\n",
	  	mpi_req->handle_type);
	  mxmpi_abort(1);
	}

	mpi_req->chandle.is_complete = 1;

      /* not active, this request has been freed, and no one cares... */
      } else {
        if (mpi_req->handle_type == MPIR_SEND) {
mpi_req->shandle.mx_req_handle = 0; /* XXX debug */
	  MPIR_FORGET_SEND(&mpi_req->shandle);
	  MPID_SendFree(mpi_req);
        } else if (mpi_req->handle_type == MPIR_RECV) {
mpi_req->rhandle.mx_req_handle = 0; /* XXX debug */
	  MPID_RecvFree(mpi_req);
	} else if (mpi_req->handle_type != MPIR_PERSISTENT_SEND &&
	           mpi_req->handle_type != MPIR_PERSISTENT_RECV) {
	  fprintf(stderr, "Mysterious free handle_type = %d\n",
	  	mpi_req->handle_type);
	  mxmpi_abort(1);
	}
      }
        
    } else {
      return found;
    }
  }

#if 0
    MPID_Device *dev;
    int found = 0;
    int lerr;

    DEBUG_PRINT_MSG( "Starting DeviceCheck")
   if (MPID_devset->ndev_list == 1) {
	dev = MPID_devset->dev_list;
	lerr = (*dev->check_device)( dev, is_blocking );
	return (lerr == 0) ? 1 : lerr;
    }
    else {
        DEBUG_PRINT_MSG( "Entering while !found" );
        while (!found) {
	    dev = MPID_devset->dev_list;
	    while (dev) {
		lerr =  (*dev->check_device)( dev, MPID_NOTBLOCKING );
		found |= (lerr == 0);
		dev = dev->next;
	    }
	    if (is_blocking == MPID_NOTBLOCKING) {
		break;
	    }
	}
	DEBUG_PRINT_MSG( "Leaving while !found" );
    }
    DEBUG_PRINT_MSG( "Exiting DeviceCheck")
    return (found) ? found : -1;
#endif
}

int MPID_Complete_pending()
{
  MXMPI_NOTIMPL();
#if 0
    MPID_Device *dev;
    int         lerr;

    DEBUG_PRINT_MSG( "Starting Complete_pending")
    if (MPID_devset->ndev_list == 1) {
	dev = MPID_devset->dev_list;
	while (MPID_n_pending > 0) {
	    lerr = (*dev->check_device)( dev, MPID_BLOCKING );
	    if (lerr > 0) {
		return lerr;
	    }
	}
    }
    else {
	while (MPID_n_pending > 0) {
	    dev = MPID_devset->dev_list;
	    while (dev) {
		lerr = (*dev->check_device)( dev, MPID_NOTBLOCKING );
		if (lerr > 0) {
		    return lerr;
		}
		dev = dev->next;
	    }
	}
    }
    DEBUG_PRINT_MSG( "Exiting Complete_pending")
    return MPI_SUCCESS;
#endif
}

void MPID_SetPktSize( int len )
{
  MXMPI_NOTIMPL();
#if 0
    MPID_Short_len = len;
#endif
}

/*
  Perhaps this should be a util function
 */
int MPID_WaitForCompleteSend( 
	MPIR_SHANDLE *request)
{
  MXMPI_NOTIMPL();
#if 0
    while (!request->is_complete)
	MPID_DeviceCheck( MPID_BLOCKING );
    return MPI_SUCCESS;
#endif
}

int MPID_WaitForCompleteRecv( 
	MPIR_RHANDLE *request)
{
  MXMPI_NOTIMPL();
#if 0
    while (!request->is_complete)
	MPID_DeviceCheck( MPID_BLOCKING );
#endif
    return MPI_SUCCESS;
}

void MPID_Version_name( char *name )
{
  MXMPI_NOTIMPL();
#if 0
    sprintf( name, "ADI version %4.2f - transport %s", MPIDPATCHLEVEL, 
	     MPIDTRANSPORT );
#endif
}

void
mxmpi_abort(int code)
{
  static int aborted = 0;
  int sockfd, count, i;
  char buffer[32];
  
  mx_finalize();

  if (aborted == 0)
    {
      if (mxmpi.mpd)
	{
	  abort();
	}
      else
	{
	  fflush (stderr);
	  fflush (stdout);
	  aborted = 1;
      
	  /* get a socket */
	  sockfd = socket (AF_INET, SOCK_STREAM, 0);
	  if (sockfd < 0)
	    {
	      fprintf (stderr, "[%d] Error: Unable to open a socket !\n", 
		       MPID_MyWorldRank);
	      exit (-1);
	    }
	  
	  /* connect to the master */
	  i = 10;
	  while (connect (sockfd, (struct sockaddr *) (&(mxmpi.master_addr)), 
			  sizeof (mxmpi.master_addr)) < 0)
	    {
#ifdef WIN32
	      Sleep (1);
#else
	      sleep(1); 
#endif
	      if (i == 0)
		{
		  fprintf (stderr, "[%d] Error: Unable to connect to "
			   "the master to abort !\n", MPID_MyWorldRank);
		  exit (-1);
		}
	      i--;
	    }
	  
	  /* send the abort message */
	  count = 0;
	  sprintf (buffer, "<<<ABORT_%d_ABORT>>>", mxmpi.filter_value);
	  while (count < strlen (buffer))
	    {
	      i = write (sockfd, &(buffer[count]), strlen (buffer) - count);
	      if (i < 0)
		{
		  fprintf (stderr, "[%d] Error: write to socket failed !\n", 
			   MPID_MyWorldRank);
		  exit (-1);;
		}
	      count += i;
	    }
	  
	  close (sockfd);
	  exit (-1);
	}
    }
}
