/*************************************************************************
 * Myricom MPICH-MX ch_mx backend                                        *
 * Copyright (c) 2003 by Myricom, Inc.                                   *
 * All rights reserved.                                                  *
 *************************************************************************/

#ifndef _mxmpi_h
#define _mxmpi_h

#include <assert.h>

#include "mpichconf.h"
#include "mpich-mpid.h"

/*  For functions not yet implemented */
#define MXMPI_NOTIMPL() \
	fprintf(stderr, "Not yet implemented at %s:%s\n", __FILE__, __LINE__);\
	exit(1); 

#define MXMPI_VERSION 0x01000000

/* tunable parameters */
#define MXMPI_EAGER_SIZE_DEFAULT 1024*16-96
#define MXMPI_MAX_PUT_LENGTH 1024*1024
#define MXMPI_INITIAL_DMA_SEND_BUF 256
#define MXMPI_MAX_RECV_TOKENS 1000
#define MXMPI_MAX_PUT_IN_PIPE 32
#define MXMPI_ENABLE_REG_CACHE 1

#define MXMPI_MAX_MX_PORTS 16
#define MXMPI_MAX_MX_BOARDS 8
#define MXMPI_INIT_TIMEOUT (3*60*1000) /* 3 minutes */
#define MXMPI_MAX_DELAYED_SEND_QUEUE_SIZE 8*1024*1024
#define MXMPI_MALLOC_HOOK_CHECK_SIZE 4*1024*1024

/* 8 MB for unlucky Solaris users (temporarly, 
   registration is coming on Solaris 8 soon) */
#define MXMPI_BOUNCE_BUFFERS 128

#undef MXMPI_PROGRESSION
#define MXMPI_PROGRESSION 0

/* debug */
#define MXMPI_DEBUG_CHECKSUM 0
#define MXMPI_DEBUG_ASSERT 0
#define MXMPI_DEBUG_REG_CACHE 0
#define MXMPI_DEBUG_FIFO_SEND 0
#define MXMPI_DEBUG_MX_SEND 0
#define MXMPI_DEBUG_MX_RECV 0
#define MXMPI_DEBUG_REGISTRATION 0
#define MXMPI_DEBUG_DMA_MEMORY 0
#define MXMPI_DEBUG_OUTPUT_FILE 0

/* MX keys */
#define MXMPI_PKT_MX_KEY 30

/* misc */
#define MXMPI_DMA_SEND_BUF 2
#define MXMPI_MALLOC_SEND_BUF 4
#define MXMPI_PUT_DATA 8

/* magic numbers for MX */
#define MXMPI_GLOBAL_MAGIC 0x12345678
#define MXMPI_LOCAL_MAGIC 0x87654321

#include <sys/types.h>
#include <netinet/in.h>

#include "myriexpress.h"
#include "mx_auto_config.h"
#include "mpid.h"
#include "mxmpi_debug_checksum.h"
#include "req.h"
#include "dev.h"

#include <netinet/in.h>

struct mxmpi_send_buf
{
  unsigned int type;
  unsigned int dest;
  unsigned int length;
  unsigned int register_length;
  void * source_ptr;
  void * target_ptr;
  MPIR_SHANDLE * shandle;
  struct mxmpi_send_buf * next;
  unsigned int buffer; 
};

/*
 * Receive modes
 */
enum {
  MXMPI_RXMODE_BLOCKING,
  MXMPI_RXMODE_POLLING
};

struct mxmpi_var
{
  mx_endpoint_t my_endpoint;
  unsigned int send_tokens;
  unsigned int max_send_tokens;
  unsigned int recv_tokens;
  unsigned int max_recv_tokens;
  unsigned int send_buf_fifo_queued;
  unsigned int send_dma_buf_pool_remaining;
  unsigned int malloc_send_buf_allocated;
  unsigned int pending_put_callback;
  unsigned int pending_send_callback;
  unsigned int unexpected_short;
  unsigned int eager_size;
  unsigned int shmem;
  unsigned int magic;
  unsigned int mpd;
  struct sockaddr_in master_addr;
  FILE * debug_output_filedesc;
  unsigned int mx_receive_mode;

  struct mxmpi_send_buf * send_buf_fifo_head;
  struct mxmpi_send_buf * send_buf_fifo_tail;
  struct mxmpi_send_buf * send_dma_buf_free;

  mx_endpoint_t *endpoints;     /* endpoints */
  char **host_names;           /* names of machines */
  char **exec_names;           /* names of executables */
  unsigned int *mpi_pids;      /* PID of remote processes */
  unsigned int *pending_sends; /* pending sends (timeout) */
  unsigned int *dropped_sends; /* dropped sends (timeout) */
};


extern struct mxmpi_var mxmpi;
extern mx_port_t mxmpi_mx_port;

unsigned long mxmpi_use_interval(unsigned long, unsigned int);
void mxmpi_unuse_interval(unsigned long, unsigned int);
void mxmpi_clear_interval(unsigned long, unsigned int);
void mxmpi_clear_all_intervals(void);
void mxmpi_regcache_init(void);

void * mxmpi_allocate_packet(unsigned int, struct mxmpi_send_buf **);
void mxmpi_send_packet(struct mxmpi_send_buf *, unsigned int);
void mxmpi_queue_packet_register(struct mxmpi_send_buf *, void *, 
				unsigned int, unsigned int);
void mxmpi_put_data(MPIR_SHANDLE *, unsigned int, unsigned int,
		   void *, void *);
void mxmpi_init(int *, char ***);
void mxmpi_finish(MPID_Device *);
int MPID_CH_Check_incoming(MPID_Device *, MPID_BLOCKING_TYPE);
void mxmpi_abort(int);
void mxmpi_malloc_assert(void *, char *, char *);

#define MXMPI_DRAIN_MX_INCOMING

#if MXMPI_DEBUG_ASSERT
#define mxmpi_debug_assert assert
#else
#define mxmpi_debug_assert(a) 
#endif


#include "mxmpi_noreg.h"
#include "mxmpi_debug_registration.h"
#include "mxmpi_debug_regcache.h"
#include "mxmpi_debug_dma_memory.h"
#include "mxmpi_debug_mx_send_recv.h"
#include "mxmpi_debug_fifo_send.h"

#endif /* mxmpi_h */
