/*************************************************************************
 * Myricom MPICH-MX ch_mx backend                                        *
 * Copyright (c) 2003 by Myricom, Inc.                                   *
 * All rights reserved.                                                  *
 *************************************************************************/

#ifdef WIN32
/* Windows doesn't use the aux funtions. */
int foo;
#else
#include <unistd.h>
#include <unistd.h>
#include <sys/mman.h>

#include "mxmpi.h"

#if !defined(__linux__) && !defined(__APPLE__)
/* loic: at least on 64bits archs, it is important that sbrk
   has the right prototype sbrk altough quite UNIX universal
   is a non-official function, so might not be in headers
   if this definition conflicts with yours, remove the OS
   from the def condition
 */
void *sbrk();
#endif

#if MX_OS_LINUX
#include <syscall.h>

int munmap(void *start, size_t length)
{
  mxmpi.malloc_hook_flag = 1;

  MXMPI_DEBUG_REG_CACHE_PRINT2("munmap (stub)", "start", start, "length", length);
  MXMPI_DEBUG_REGISTRATION_CHECK_ALIGN("munmap (stub)", (unsigned long)(start),
                                      length);
  mxmpi_clear_interval((unsigned long)start, length);
  syscall(__NR_munmap,start,length);
}

int __munmap(void *start, size_t length)
{
  return munmap(start,length);
}

#endif

void *mxmpi_sbrk(int inc)
{
  mxmpi.malloc_hook_flag = 1;
      
  if (inc < 0)
    {
      long oldp = (long)sbrk(0);
      
      MXMPI_DEBUG_REG_CACHE_PRINT2("mxmpi_sbrk", "oldp", oldp, "inc", inc);
      MXMPI_DEBUG_REGISTRATION_CHECK_ALIGN("mxmpi_sbrk", 
					  (unsigned long)(oldp+inc), (-inc));
      mxmpi_clear_interval((unsigned long)(oldp+inc), -inc);
    }
  return sbrk(inc);
}


int mxmpi_munmap(void *start, size_t length)
{
  mxmpi.malloc_hook_flag = 1;

  MXMPI_DEBUG_REG_CACHE_PRINT2("mxmpi_munmap", "start", start, "length", length);
  MXMPI_DEBUG_REGISTRATION_CHECK_ALIGN("mxmpi_munmap", (unsigned long)(start),
				      length);
  mxmpi_clear_interval((unsigned long)start, length);
  return munmap(start, length);
}
#endif
