/*************************************************************************
 * Myricom MPICH-MX ch_mx backend                                        *
 * Copyright (c) 2003 by Myricom, Inc.                                   *
 * All rights reserved.                                                  *
 *************************************************************************/

#include "mxmpi.h"


static mxmpi_bounce_buffer *bounce_free = NULL;
static mxmpi_bounce_buffer *bounce_used = NULL;
static mxmpi_bounce_buffer bounce_buffers[MXMPI_BOUNCE_BUFFERS];

void 
mxmpi_bounce_buffer_init(void)
{
  mxmpi_bounce_buffer *previous_buffer = NULL;
  unsigned int i;

  mxmpi_debug_assert(MX_PAGE_LEN != 0);
  if ((bounce_free == NULL) && (bounce_used == NULL))
    {
      mx_bzero(bounce_buffers, 
	       MXMPI_BOUNCE_BUFFERS * sizeof(mxmpi_bounce_buffer));
      
      for (i=0; i<MXMPI_BOUNCE_BUFFERS; i++)
	{
	  bounce_buffers[i].status = MXMPI_BOUNCE_SEGMENT_FREE;
	  bounce_buffers[i].next = previous_buffer;
	  bounce_buffers[i].addr = 
	    (unsigned long) mxmpi_dma_alloc(MXMPI_MAX_PUT_LENGTH);
	  mxmpi_malloc_assert((void *)(bounce_buffers[i].addr),
			     "mxmpi_bounce_buffer_init",
			     "mxmpi_dma_alloc: bounce buffer");
	  MXMPI_DEBUG_DMA_MEMORY_ACQUIRE(MXMPI_MAX_PUT_LENGTH);
	  MXMPI_DEBUG_DMA_MEMORY_USE(MXMPI_MAX_PUT_LENGTH);
	  previous_buffer = &(bounce_buffers[i]);
	}
      
      bounce_free = &(bounce_buffers[MXMPI_BOUNCE_BUFFERS-1]);
    }
}


void
mxmpi_bounce_buffer_finish(void)
{
  unsigned int i;
  
  if ((bounce_free != NULL) || (bounce_used != NULL))
    {
      for (i=0; i<MXMPI_BOUNCE_BUFFERS; i++)
	{
	  mxmpi_dma_free((void *)(bounce_buffers[i].addr));
	  MXMPI_DEBUG_DMA_MEMORY_RELEASE(MXMPI_MAX_PUT_LENGTH);
	  MXMPI_DEBUG_DMA_MEMORY_UNUSE(MXMPI_MAX_PUT_LENGTH);
	}
      mx_bzero(bounce_buffers, 
	       MXMPI_BOUNCE_BUFFERS * sizeof(mxmpi_bounce_buffer));
      bounce_free = NULL;
      bounce_used = NULL;
    }
}


unsigned long 
mxmpi_allocate_bounce_buffer(unsigned long data_addr,
			    unsigned int length)
{
  mxmpi_bounce_buffer * bounce_buffer;
  
  MXMPI_DEBUG_REG_CACHE_PRINT1("Allocate bounce buffer", "length", length);
  mxmpi_debug_assert(bounce_buffers[0].addr != 0);
  
  if (bounce_free != NULL)
    {
      mxmpi_debug_assert(bounce_free->status == MXMPI_BOUNCE_SEGMENT_FREE);
      bounce_buffer = bounce_free;
      bounce_buffer->data_addr = data_addr;
      bounce_buffer->length = length;
      bounce_buffer->status = MXMPI_BOUNCE_SEGMENT_USED;
      bounce_free = bounce_buffer->next;
      bounce_buffer->next = bounce_used;
      bounce_used = bounce_buffer;
      
      MXMPI_DEBUG_REG_CACHE_PRINT2("Allocate bounce buffer",
				  "data_addr", data_addr,
				  "bounce_addr", bounce_buffer->addr); 
      return (bounce_buffer->addr);
    }
  else
    {	
      MXMPI_DEBUG_REG_CACHE_PRINT0("Allocate bounce buffer: no luck");
      return 0;
    }

}


void mxmpi_free_bounce_buffer(unsigned long data_addr,
			     unsigned int length)
{
  mxmpi_bounce_buffer * bounce_buffer;
  mxmpi_bounce_buffer * previous_bounce_buffer;
  
  MXMPI_DEBUG_REG_CACHE_PRINT2("Free bounce buffer", "data_addr",
			      data_addr, "length", length);
  
  previous_bounce_buffer = NULL;
  bounce_buffer = bounce_used;
  while (bounce_buffer != NULL)
    {
      if (bounce_buffer->data_addr == data_addr)
	{
	  mxmpi_debug_assert(bounce_buffer->length == length);
	  mxmpi_debug_assert(bounce_buffer->status == MXMPI_BOUNCE_SEGMENT_USED);
	  bounce_buffer->status = MXMPI_BOUNCE_SEGMENT_FREE;
	  if (previous_bounce_buffer != NULL)
	    {
	      previous_bounce_buffer->next = bounce_buffer->next;
	    }
	  else
	    {
	      bounce_used = bounce_buffer->next;
	    }
	  bounce_buffer->next = bounce_free;
	  bounce_free = bounce_buffer;
	  return;
	}
      else
	{
	  previous_bounce_buffer = bounce_buffer;
	  bounce_buffer = bounce_buffer->next;
	}
    }
  mxmpi_debug_assert(0);
}
