/*************************************************************************
 * Myricom MPICH-MX ch_mx backend                                        *
 * Copyright (c) 2003 by Myricom, Inc.                                   *
 * All rights reserved.                                                  *
 *************************************************************************/


#include <stdarg.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifdef WIN32
#include <windows.h>
#else
#include <sys/time.h>
#include <unistd.h>
#endif

#include "mxmpi.h"

#include "mpid.h"
#include "mpiddev.h"
#include "mpid_bind.h"
#include "packets.h"
#include "queue.h"
#include "reqalloc.h"

#include "myriexpress.h"


struct mxmpi_var  mxmpi;
mx_port_t mxmpi_mx_port = NULL;
static unsigned int callback_flag;


static void mxmpi_flush_fifo_send(void);
static mx_callback_t mxmpi_send_packet_callback;
static mx_callback_t mxmpi_put_data_callback;
int MPID_CH_Rndvn_ok_to_send(MPIR_RHANDLE *);
void MPD_Abort (int);


inline void 
mxmpi_malloc_assert(void * ptr, char * fct, char * msg)
{
  if (ptr == NULL)
    {
      fprintf(stderr, "[%d]: alloc failed, not enough memory (Fatal Error)\n"
	      "Context: <(%s) %s>\n", MPID_MyWorldRank, fct, msg);
      mxmpi_abort (0);
    }
}


#if 0  /* reese XXX */
void * 
mxmpi_dma_alloc(unsigned int length)
{
  void * ptr;
  
  ptr = mx_dma_malloc(mxmpi_mx_port, length);
  MXMPI_DEBUG_DMA_MEMORY_ALLOC(ptr, length);
  return(ptr);
}


void 
mxmpi_dma_free(void *ptr)
{
  MXMPI_DEBUG_DMA_MEMORY_FREE(ptr);
  mx_dma_free(mxmpi_mx_port, ptr);
}
#endif

static void 
mxmpi_drop_send_packet_callback (mx_port_t port,
				void *context,
				mx_status_t *status)
{
  struct mxmpi_send_buf * mxmpi_send_buf_ptr;
  mx_segment_t seg;
  mx_handle_t handle;
  
  mxmpi_send_buf_ptr = (struct mxmpi_send_buf *)context;
  switch (status->code)
    {
    case MX_SUCCESS:  
      mxmpi.dropped_sends[mxmpi_send_buf_ptr->dest]--;
      seg.segment_ptr = (void *)(&(mxmpi_send_buf_ptr->buffer));
      seg.segment_length = mxmpi_send_buf_ptr->length;
      mx_post_send(port, /* reese XXX */
	           &seg,
		   1,
		   mxmpi.endpoints[mxmpi_send_buf_ptr->dest],
		   MXMPI_PKT_MX_KEY,
		   mxmpi_send_packet_callback,
		   context,
		   &handle);
      break;
      
    default:
      fprintf(stderr,
	      "FATAL ERROR on MPI node %d (mx_endpoint %s): Drop callback "
	      "status is unexpected (%d)\n",
	      MPID_MyWorldRank, 
              endpoint_to_text(mxmpi.endpoints[MPID_MyWorldRank]),
	      status->code);
      mxmpi_abort (0);
    }
}

static void 
mxmpi_drop_put_data_callback (struct mx_port *port,
			     void *context,
			     mx_status_t *status)
{
  struct mxmpi_send_buf * mxmpi_send_buf_ptr;
  
  mxmpi_send_buf_ptr = (struct mxmpi_send_buf *)context;
  switch (status)
    {
    case MX_SUCCESS:  
      mxmpi.dropped_sends[mxmpi_send_buf_ptr->dest]--;
reese XXX
      mx_directed_send_with_callback 
	(port, (void *)(mxmpi_send_buf_ptr->source_ptr),
	 (mx_remote_ptr_t)(mx_up_t)(mxmpi_send_buf_ptr->target_ptr),
	 mxmpi_send_buf_ptr->length,
	 MX_LOW_PRIORITY,
	 mxmpi.node_ids[mxmpi_send_buf_ptr->dest], 
	 mxmpi.port_ids[mxmpi_send_buf_ptr->dest],
	 mxmpi_put_data_callback, context); 
      break;
      
    default:
      fprintf(stderr,
	      "FATAL ERROR on MPI node %d (mx_id %d): Drop put callback "
	      "status is unexpected (%d)\n", MPID_MyWorldRank, 
              mxmpi.node_ids[MPID_MyWorldRank], status);
      mxmpi_abort (0);
    }
}


mx_inline int 
mxmpi_check_send_status(mx_status_t status, int dest, int type, 
		       void * context)
{
  switch (status.code)
    {
    case MX_SUCCESS:
      return 1;
      
    default:
      fprintf(stderr,
	      "FATAL ERROR on MPI node %d (mx_endpoint %s): MX send to "
              "MPI node %d (mx_endpoint %s) failed for unknown reason (%d)\n",
              MPID_MyWorldRank,
	      mx_endpoint_to_text(mxmpi.endpoints[MPID_MyWorldRank]),
	      dest,
	      mx_endpoint_to_text(mxmpi.endpoints[dest]),
	      status.code); 
      return 0;
    }
}


static void 
mxmpi_send_packet_callback(struct mx_port * port,
			  void * context,
			  mx_status_t status)
{
  unsigned int err_code;
  struct mxmpi_send_buf * mxmpi_send_buf_ptr;
  
  mxmpi_send_buf_ptr = (struct mxmpi_send_buf *)context;
  err_code = mxmpi_check_send_status(status, mxmpi_send_buf_ptr->dest, 1, 
				    context);
  
  switch (err_code)
    {
    case 1:
      MXMPI_DEBUG_MX_SEND_FINISH (0, (void *)(&(mxmpi_send_buf_ptr->buffer)),
				 mxmpi_send_buf_ptr->length,
				 mxmpi_send_buf_ptr->dest, context);
      
      mxmpi_debug_assert(mxmpi_send_buf_ptr->type == MXMPI_DMA_SEND_BUF);
      mxmpi_send_buf_ptr->next = mxmpi.send_dma_buf_free;
      mxmpi.send_dma_buf_free = mxmpi_send_buf_ptr;
      mxmpi.send_dma_buf_pool_remaining++;
      
      callback_flag = 1;
      mxmpi.pending_sends[mxmpi_send_buf_ptr->dest]--;
      mxmpi.pending_send_callback--;
      mxmpi_debug_assert(mxmpi.send_tokens < mxmpi.max_send_tokens);
      mxmpi.send_tokens++;
      
      if (mxmpi.send_buf_fifo_head != NULL)
	{
	  mxmpi_flush_fifo_send();
	}
      break;
  
    case 2:
      break;

    case 0:
      fprintf(stderr, "Small/Ctrl message completion error!\n");
      mxmpi_abort (0);
      break;
      
    default:
      fprintf(stderr, "Bad status code in send callback!\n");
      mxmpi_abort (0);
    }
}


static void 
mxmpi_put_data_callback(struct mx_port * port,
		       void * context,
		       mx_status_t status)
{
  unsigned int err_code;
  unsigned long chunk_size;
  struct mxmpi_send_buf *mxmpi_send_buf_ptr;
  MPIR_SHANDLE *shandle;

  mxmpi_send_buf_ptr = (struct mxmpi_send_buf *)context;
  shandle = mxmpi_send_buf_ptr->shandle;
  mxmpi_debug_assert (mxmpi_send_buf_ptr->dest == shandle->partner);
  err_code = mxmpi_check_send_status(status, mxmpi_send_buf_ptr->dest, 2, 
				    context);
  
  switch (err_code)
    {
    case 1:
      chunk_size = (shandle->bytes_as_contig - shandle->mx.current_done);
      if (chunk_size > MXMPI_MAX_PUT_LENGTH)
	{
	  chunk_size = MXMPI_MAX_PUT_LENGTH;
	}
      mxmpi_debug_assert (chunk_size == mxmpi_send_buf_ptr->length);
      mxmpi_debug_assert (((unsigned long)shandle->start 
			  + shandle->mx.current_done) == 
			 (unsigned long) (mxmpi_send_buf_ptr->source_ptr));
  
      MXMPI_DEBUG_MX_SEND_FINISH(1, (void *)(mxmpi_send_buf_ptr->source_ptr),
				mxmpi_send_buf_ptr->length,
				mxmpi_send_buf_ptr->dest, context);
  
#if MX_DISABLE_REGISTRATION
      mxmpi_free_bounce_buffer((unsigned long) (mxmpi_send_buf_ptr->source_ptr),
			      mxmpi_send_buf_ptr->length);
#else
      mxmpi_unuse_interval((unsigned long) (mxmpi_send_buf_ptr->source_ptr),
			  mxmpi_send_buf_ptr->length);
#endif
      
      shandle->mx.current_done += chunk_size;
      
      mxmpi_debug_assert(shandle->mx.current_done <= shandle->bytes_as_contig);
      if (shandle->mx.current_done == shandle->bytes_as_contig)
	{
	  shandle->is_complete = 1;
	  if (shandle->finish)
	    {
	      (shandle->finish)(shandle);
	    }
	}
      mxmpi.pending_sends[mxmpi_send_buf_ptr->dest]--;
      free(mxmpi_send_buf_ptr);
      
      callback_flag = 1;
      mxmpi.pending_put_callback--;
      mxmpi_debug_assert(mxmpi.send_tokens < mxmpi.max_send_tokens);
      mxmpi.send_tokens++;
      
      if (mxmpi.send_buf_fifo_head != NULL)
	{
	  mxmpi_flush_fifo_send();
	}
      break;
  
    case 2:
      break;
      
    case 0:
      fprintf(stderr, "Large message completion error !\n");
      mxmpi_abort (0);
      break;
        
    default:
      fprintf(stderr, "Bad status code in send callback!\n");
      mxmpi_abort (0);
    }
}


static void 
mxmpi_flush_fifo_send (void)
{
  struct mxmpi_send_buf *mxmpi_send_buf_ptr;
  struct mxmpi_send_buf *mxmpi_new_send_buf_ptr;
  
  mxmpi_debug_assert (mxmpi.send_buf_fifo_head != NULL);
  
  while (mxmpi.send_tokens > 0)
    {
      mxmpi_debug_assert(mxmpi.send_buf_fifo_queued > 0);
      mxmpi_send_buf_ptr = mxmpi.send_buf_fifo_head;

      mxmpi_debug_assert(mxmpi_send_buf_ptr->dest < MPID_MyWorldSize);
      if (mxmpi.dropped_sends[mxmpi_send_buf_ptr->dest] > 0)
	{
	  return;
	}

      if ((mxmpi.send_dma_buf_free == NULL) 
	  && (mxmpi_send_buf_ptr->type == MXMPI_MALLOC_SEND_BUF))
	{
	  mxmpi_debug_assert(mxmpi.send_dma_buf_pool_remaining == 0);
	  return;
	}
      
      if (mxmpi_send_buf_ptr->register_length > 0)
	{
#if MX_DISABLE_REGISTRATION
	  unsigned long bounce_addr, bounce_data;
	  unsigned int bounce_length;
	  
	  bounce_length = mxmpi_send_buf_ptr->register_length;
	  bounce_data = (unsigned long)mxmpi_send_buf_ptr->source_ptr;
	  bounce_addr = mxmpi_allocate_bounce_buffer(bounce_data,
						    bounce_length);
	  
	  if (bounce_addr != 0)
	    {
	      if (mxmpi_send_buf_ptr->type == MXMPI_PUT_DATA)
		{
		  memcpy((void *)bounce_addr,
			 (void *)bounce_data,
			 bounce_length);
		  mxmpi_send_buf_ptr->source_ptr = (void *)bounce_addr;
		}
	      else
		{
		  MPID_PKT_OK_TO_SEND_T * pkt_ptr;

		  pkt_ptr = ((MPID_PKT_OK_TO_SEND_T *)
			     (&(mxmpi_send_buf_ptr->buffer)));
		  mxmpi_debug_assert(pkt_ptr->mode 
				    == MPID_PKT_OK_TO_SEND);
		  MPID_AINT_SET(pkt_ptr->target_ptr,
				(void *)bounce_addr);
		  
		  MXMPI_DEBUG_CHECKSUM_COMPUTE((struct pkt_cksum *)pkt_ptr,
					      (void *)
					      ((unsigned long)pkt_ptr 
					       + sizeof(struct pkt_cksum)),
					      mxmpi_send_buf_ptr->length 
					      - sizeof(struct pkt_cksum));
		}
	    }
	  else
	    {
	      return;
	    }
#else
	  unsigned int registered;
	  
	  registered = mxmpi_use_interval((unsigned long)
					 (mxmpi_send_buf_ptr->source_ptr),
					 mxmpi_send_buf_ptr->register_length);
	  
	  if (registered < mxmpi_send_buf_ptr->register_length)
	    {
	      mxmpi_unuse_interval((unsigned long)
				  (mxmpi_send_buf_ptr->source_ptr), 
				  registered);
	      return;
	    }
#endif
	  
	  mxmpi_send_buf_ptr->register_length = 0;
	}
      
      if (mxmpi_send_buf_ptr->type == MXMPI_MALLOC_SEND_BUF)
        {
	  mxmpi_debug_assert(mxmpi.send_dma_buf_free != NULL);
	  mxmpi_debug_assert(mxmpi.send_dma_buf_pool_remaining > 0);
	  mxmpi_debug_assert(mxmpi_send_buf_ptr->target_ptr == NULL);
	  mxmpi_debug_assert(mxmpi_send_buf_ptr->length > 0);
	  
	  mxmpi_new_send_buf_ptr = mxmpi.send_dma_buf_free;
	  mxmpi.send_dma_buf_free = mxmpi_new_send_buf_ptr->next;
	  mxmpi.send_dma_buf_pool_remaining--;
	  memcpy(mxmpi_new_send_buf_ptr, mxmpi_send_buf_ptr, 
		 sizeof(struct mxmpi_send_buf) - sizeof (unsigned int)
		 + mxmpi_send_buf_ptr->length);
	  
          mxmpi.send_buf_fifo_head = mxmpi_new_send_buf_ptr;
          if (mxmpi.send_buf_fifo_tail == mxmpi_send_buf_ptr) 
            {
              mxmpi.send_buf_fifo_tail = mxmpi_new_send_buf_ptr;
            } 
	  
	  mxmpi_debug_assert(mxmpi.malloc_send_buf_allocated > 0);
	  mxmpi.malloc_send_buf_allocated -= mxmpi_send_buf_ptr->length;
          free(mxmpi_send_buf_ptr);
	  
	  mxmpi_send_buf_ptr = mxmpi_new_send_buf_ptr;
	  mxmpi_send_buf_ptr->type = MXMPI_DMA_SEND_BUF;
	} 

      switch(mxmpi_send_buf_ptr->type) 
        { 
	case MXMPI_DMA_SEND_BUF:
        
          MXMPI_DEBUG_MX_SEND_PRINT(0, 1, 
				   (void *)(&(mxmpi_send_buf_ptr->buffer)),
				   mxmpi_send_buf_ptr->length,
				   mxmpi_send_buf_ptr->dest, NULL,
                                   (void *)mxmpi_send_buf_ptr);
	 
          mxmpi_debug_assert(mxmpi.send_tokens > 0); 
	  mxmpi.pending_sends[mxmpi_send_buf_ptr->dest]++;
	  mx_send_with_callback(mxmpi_mx_port, 
				(void *)(&(mxmpi_send_buf_ptr->buffer)),
                                MXMPI_PKT_MX_SIZE,
                                mxmpi_send_buf_ptr->length,
				MX_LOW_PRIORITY,
                                mxmpi.node_ids[mxmpi_send_buf_ptr->dest],
                                mxmpi.port_ids[mxmpi_send_buf_ptr->dest],
                                mxmpi_send_packet_callback,
                                (void *)mxmpi_send_buf_ptr);
	  
	  mxmpi.pending_send_callback++;
	  mxmpi.send_tokens--;
	  mxmpi.send_buf_fifo_queued--;
	  MXMPI_DEBUG_FIFO_SEND_REMOVE(0, mxmpi.send_buf_fifo_queued);
	  mxmpi.send_buf_fifo_head = mxmpi_send_buf_ptr->next;
	  break;
	  
	case MXMPI_PUT_DATA:
	  mxmpi_debug_assert(mxmpi_send_buf_ptr->source_ptr != NULL);
	  mxmpi_debug_assert(mxmpi_send_buf_ptr->target_ptr != NULL);
	  mxmpi_debug_assert(mxmpi_send_buf_ptr->shandle != NULL);
	  mxmpi_debug_assert(mxmpi_send_buf_ptr->length > 0);
          mxmpi_debug_assert(mxmpi_send_buf_ptr->buffer == 99);
	  
	  MXMPI_DEBUG_MX_SEND_PRINT(1, 1, 
				   mxmpi_send_buf_ptr->source_ptr,
				   mxmpi_send_buf_ptr->length,
				   mxmpi_send_buf_ptr->dest,
				   mxmpi_send_buf_ptr->target_ptr,
                                   (void *)(mxmpi_send_buf_ptr->shandle));
	  
	  MXMPI_DEBUG_REGISTRATION_CHECK_SEGMENT((unsigned long)
						mxmpi_send_buf_ptr->source_ptr,
						mxmpi_send_buf_ptr->length, 
						"PUT from FIFO");
	  
	  mxmpi_debug_assert(mxmpi.send_tokens > 0);
          mxmpi.pending_sends[mxmpi_send_buf_ptr->dest]++;
	  mx_directed_send_with_callback
	    (mxmpi_mx_port,
	     mxmpi_send_buf_ptr->source_ptr,
	     (mx_remote_ptr_t)(mx_up_t)mxmpi_send_buf_ptr->target_ptr,
	     mxmpi_send_buf_ptr->length,
	     MX_LOW_PRIORITY,
	     mxmpi.node_ids[mxmpi_send_buf_ptr->dest],
	     mxmpi.port_ids[mxmpi_send_buf_ptr->dest],
	     mxmpi_put_data_callback,
	     (void *)mxmpi_send_buf_ptr); 
	  
	  mxmpi.pending_put_callback++;
	  mxmpi.send_tokens--;
          mxmpi.send_buf_fifo_queued--;
	  MXMPI_DEBUG_FIFO_SEND_REMOVE(1, mxmpi.send_buf_fifo_queued);
	  mxmpi.send_buf_fifo_head = mxmpi_send_buf_ptr->next;
	  break;
	  
	default:
	  fprintf(stderr, "[%d] mxmpi_flush_fifo_send: bad send buf type\n", 
		  MPID_MyWorldRank);
	  mxmpi_abort (0);
        }

      if (mxmpi.send_buf_fifo_head == NULL)
	{
	  mxmpi_debug_assert(mxmpi.send_buf_fifo_tail == mxmpi_send_buf_ptr);
	  mxmpi.send_buf_fifo_tail = NULL;
	  mxmpi_debug_assert(mxmpi.send_buf_fifo_queued == 0);
	  return;
	}
    }
}


mx_inline void * 
mxmpi_allocate_packet(unsigned int length, struct mxmpi_send_buf ** send_buf_ptr)
{
  struct mxmpi_send_buf * mxmpi_send_buf_ptr;
  
  if (mxmpi.send_dma_buf_free != NULL)
    {
      mxmpi_debug_assert(mxmpi.send_dma_buf_pool_remaining > 0);
      mxmpi_send_buf_ptr = mxmpi.send_dma_buf_free;
      mxmpi_debug_assert(mxmpi_send_buf_ptr->type == MXMPI_DMA_SEND_BUF);
      mxmpi.send_dma_buf_free = mxmpi_send_buf_ptr->next;
      mxmpi.send_dma_buf_pool_remaining--;
    }
  else
    {
      mxmpi_debug_assert(mxmpi.send_dma_buf_pool_remaining == 0);
      mxmpi_send_buf_ptr = (struct mxmpi_send_buf *)
        malloc(length + sizeof(struct mxmpi_send_buf));
      mxmpi_malloc_assert(mxmpi_send_buf_ptr,
	                 "mxmpi_allocate_packet",
			 "malloc: send buf");
      mxmpi_send_buf_ptr->type = MXMPI_MALLOC_SEND_BUF;
      mxmpi.malloc_send_buf_allocated += length;
    }
  
  mxmpi_send_buf_ptr->dest = -1;
  mxmpi_send_buf_ptr->length = length;
  mxmpi_send_buf_ptr->register_length = 0;
  mxmpi_send_buf_ptr->source_ptr = NULL;
  mxmpi_send_buf_ptr->target_ptr = NULL;
  mxmpi_send_buf_ptr->shandle = NULL;
  mxmpi_send_buf_ptr->next = NULL;
  mxmpi_send_buf_ptr->buffer = 0;
  *send_buf_ptr = mxmpi_send_buf_ptr;
  return ((void *)(&(mxmpi_send_buf_ptr->buffer))); 
}


/*
 * Send a packet via MX...
 */
mx_inline void 
mxmpi_send_packet(struct mxmpi_send_buf * mxmpi_send_buf_ptr, unsigned int dest)
{
  void * pkt;
  mx_segment_t seg;
  
  mxmpi_debug_assert(dest < MPID_MyWorldSize);
  mxmpi_debug_assert(mxmpi_send_buf_ptr != NULL);
  
  pkt = ((void *)(&(mxmpi_send_buf_ptr->buffer)));
  mxmpi_send_buf_ptr->dest = dest;
  
  MXMPI_DEBUG_CHECKSUM_COMPUTE((struct pkt_cksum *)pkt,
                              (void *)((unsigned long)pkt 
                                       + sizeof(struct pkt_cksum)),
                              mxmpi_send_buf_ptr->length 
                              - sizeof(struct pkt_cksum));
  MXMPI_DEBUG_CHECKSUM_INFO((struct pkt_cksum *)pkt,
                           MXMPI_DEBUG_CHECKSUM_SEND
                           | mxmpi_send_buf_ptr->type);

  /* build segment to hold buffer */
  seg.segment_ptr = pkt;
  seg.segment_length = mxmpi_send_buf_ptr->length;

  rc = mx_post_send(mxmpi_mx_port, 
               &seg,
               1,
               mxmpi.endpoints[dest],
               MATCH,
               mxmpi_send_packet_callback,
               &(mxmpi_send_buf_ptr->handle));

  MXMPI_DEBUG_MX_SEND_PRINT(0, 0, pkt,
			    mxmpi_send_buf_ptr->length,
			    dest, NULL, (void *)mxmpi_send_buf_ptr);
      
  mxmpi.pending_sends[dest]++;
}


mx_inline void 
mxmpi_queue_packet_register(struct mxmpi_send_buf * mxmpi_send_buf_ptr,
			   void * reg_start,
			   unsigned int reg_length,
			   unsigned int dest)
{
  void * pkt;
  
  mxmpi_debug_assert(dest < MPID_MyWorldSize);
  
  mxmpi_send_buf_ptr->register_length = reg_length;
  mxmpi_send_buf_ptr->source_ptr = reg_start;
  mxmpi_send_buf_ptr->dest = dest;
  pkt = ((void *)(&(mxmpi_send_buf_ptr->buffer)));
  
  MXMPI_DEBUG_CHECKSUM_COMPUTE((struct pkt_cksum *)pkt,
                              (void *)((unsigned long)pkt 
                                       + sizeof(struct pkt_cksum)),
                              mxmpi_send_buf_ptr->length 
                              - sizeof(struct pkt_cksum));
  MXMPI_DEBUG_CHECKSUM_INFO((struct pkt_cksum *)pkt, 
                           MXMPI_DEBUG_CHECKSUM_QUEUE_REG
			   | mxmpi_send_buf_ptr->type);

  if (mxmpi.send_buf_fifo_head == NULL)
    {
      mxmpi_debug_assert(mxmpi.send_buf_fifo_queued == 0);
      mxmpi_debug_assert(mxmpi.send_buf_fifo_tail == NULL);
      mxmpi.send_buf_fifo_head = mxmpi_send_buf_ptr;
    }
  else
    {
      mxmpi_debug_assert(mxmpi.send_buf_fifo_tail->next == NULL);
      mxmpi.send_buf_fifo_tail->next = mxmpi_send_buf_ptr;
    }
  
  mxmpi.send_buf_fifo_tail = mxmpi_send_buf_ptr;
  mxmpi_send_buf_ptr->next = NULL;
  mxmpi.send_buf_fifo_queued++;
  MXMPI_DEBUG_FIFO_SEND_ADD(0, 1, mxmpi.send_buf_fifo_queued);
}


mx_inline void 
mxmpi_put_data(MPIR_SHANDLE * shandle,
	      unsigned int dest,
	      unsigned int length,
	      void * start,
	      void * target_ptr)
{
  struct mxmpi_send_buf * mxmpi_send_buf_ptr;
  unsigned int registered;
#if MX_DISABLE_REGISTRATION
  unsigned long bounce_addr;
#endif

  mxmpi_debug_assert(shandle != NULL);
  mxmpi_debug_assert(start != NULL);
  mxmpi_debug_assert(target_ptr != NULL);
  mxmpi_debug_assert(dest < MPID_MyWorldSize);
  mxmpi_debug_assert(length > 0);
  
  mxmpi_send_buf_ptr = 
    (struct mxmpi_send_buf *) malloc (sizeof (struct mxmpi_send_buf));
  mxmpi_malloc_assert (mxmpi_send_buf_ptr,
		      "mxmpi_put_data",
		      "malloc: send buf for out data");

  mxmpi_send_buf_ptr->type = MXMPI_PUT_DATA;
  mxmpi_send_buf_ptr->dest = dest;
  mxmpi_send_buf_ptr->length = length;
  mxmpi_send_buf_ptr->register_length = 0;
  mxmpi_send_buf_ptr->source_ptr = start;
  mxmpi_send_buf_ptr->target_ptr = target_ptr;
  mxmpi_send_buf_ptr->shandle = shandle;
  mxmpi_send_buf_ptr->buffer = 99;
  
  registered = 0;
  if ((mxmpi.send_tokens > 0) 
      && (mxmpi.send_buf_fifo_head == NULL)
      && (mxmpi.dropped_sends[dest] == 0))
    {
      mxmpi_debug_assert(mxmpi.send_buf_fifo_queued == 0);
      
#if MX_DISABLE_REGISTRATION
      bounce_addr = mxmpi_allocate_bounce_buffer((unsigned long)
						start, length);	  
      if (bounce_addr != 0)
	{
	  memcpy((void *)bounce_addr, start, length);
	  start = (void *)bounce_addr;
	  registered = length;
	}
      else
	{
	  registered = 0;
	}
#else
      registered = mxmpi_use_interval((unsigned long)start, 
				     (unsigned long)length);
#endif
      if (registered == length)
	{
	  MXMPI_DEBUG_MX_SEND_PRINT(1, 0, start, registered, 
				   dest, target_ptr, (void *)shandle);
	  MXMPI_DEBUG_REGISTRATION_CHECK_SEGMENT((unsigned long)start, 
						registered, "PUT");
	  
          mxmpi_debug_assert(mxmpi.send_tokens > 0);
          mxmpi.pending_sends[dest]++;
	  mx_directed_send_with_callback (mxmpi_mx_port, start,
					  (mx_remote_ptr_t)(mx_up_t)target_ptr,
					  registered,
					  MX_LOW_PRIORITY,
					  mxmpi.node_ids[dest],
					  mxmpi.port_ids[dest],
					  mxmpi_put_data_callback,
					  (void *)mxmpi_send_buf_ptr);
	  
	  mxmpi.pending_put_callback++;
	  mxmpi.send_tokens--;
	  return;
	}
      else
	{
#if !MX_DISABLE_REGISTRATION
	  mxmpi_unuse_interval((unsigned long)start, 
			      (unsigned long)registered);
#endif
	}
    }
  
  mxmpi_send_buf_ptr->register_length = length;
  
  if (mxmpi.send_buf_fifo_head == NULL)
    {
      mxmpi_debug_assert(mxmpi.send_buf_fifo_queued == 0);
      mxmpi_debug_assert(mxmpi.send_buf_fifo_tail == NULL);
      mxmpi.send_buf_fifo_head = mxmpi_send_buf_ptr;
    }
  else
    {
      mxmpi_debug_assert(mxmpi.send_buf_fifo_tail->next == NULL);
      mxmpi.send_buf_fifo_tail->next = mxmpi_send_buf_ptr;
    }
  
  mxmpi.send_buf_fifo_tail = mxmpi_send_buf_ptr;
  mxmpi_send_buf_ptr->next = NULL;
  mxmpi.send_buf_fifo_queued++;
  MXMPI_DEBUG_FIFO_SEND_ADD(1, 0, mxmpi.send_buf_fifo_queued);
}


mx_inline int 
mxmpi_packet_recv_event(MPID_Device * dev, void * buf)
{
  int from_grank, is_posted, err;
  MPIR_RHANDLE * rhandle;
  MPID_PKT_T * pkt_ptr;
 
  pkt_ptr = buf;
  err = MPI_SUCCESS;
  from_grank = pkt_ptr->head.src;
  
  MXMPI_DEBUG_CHECKSUM_CHECK("Small message",
			    (void *)((unsigned long)buf 
                                     + sizeof(struct pkt_cksum)),
                            from_grank, (struct pkt_cksum *)buf);
  DEBUG_PRINT_PKT("R received message", pkt_ptr);
    
  /* Separate the incoming messages from control messages */
  if (MPID_PKT_IS_MSG(pkt_ptr->head.mode))
    {
      DEBUG_PRINT_RECV_PKT("R rcvd msg", pkt_ptr);
      
      /* Is the message expected or not? 
	 This routine RETURNS a rhandle, creating one if the message 
	 is unexpected (is_posted == 0) */
      MPID_Msg_arrived(pkt_ptr->head.lrank,
		       pkt_ptr->head.tag,
		       pkt_ptr->head.context_id, 
		       &rhandle, &is_posted);
      
      MPID_DO_HETERO(rhandle->msgrep = (MPID_Msgrep_t)pkt_ptr->head.msgrep);
#ifdef MPID_DEBUG_ALL
      if (MPID_DebugFlag)
	{
	  FPRINTF(MPID_DEBUG_FILE, "[%d]R msg was %s (%s:%d)\n", 
		  MPID_MyWorldRank, is_posted ? "posted" : "unexpected",
		  __FILE__, __LINE__);
	}
#endif
      
      if (is_posted)
	{
	  switch (pkt_ptr->head.mode)
	    {
	    case MPID_PKT_SHORT:
	      MXMPI_DEBUG_MX_RECV_PRINT("expected", "Short msg", from_grank);
	      DEBUG_TEST_FCN(dev->short_msg->recv,"dev->short->recv");
	      err = (*dev->short_msg->recv)(rhandle, from_grank, pkt_ptr);
	      break;
	      
	    case MPID_PKT_REQUEST_SEND:
	      MXMPI_DEBUG_MX_RECV_PRINT("expected", "Request_to_send", 
				       from_grank);
	      DEBUG_TEST_FCN(dev->rndv->irecv,"dev->rndv->irecv");
	      err = (*dev->rndv->irecv)(rhandle, from_grank, pkt_ptr);
	      break;
	      
	    default:
	      fprintf(stderr, "[%d] Internal error: msg packet discarded "
		      "(%s:%d)\n", MPID_MyWorldRank, __FILE__, __LINE__);
	    }
	}
      else
	{
	  /* unexpected */
	  switch (pkt_ptr->head.mode)
	    {
	    case MPID_PKT_SHORT:
	      MXMPI_DEBUG_MX_RECV_PRINT("unexpected", "Short msg", from_grank);
	      DEBUG_TEST_FCN(dev->short_msg->unex,"dev->short->unex");
	      err = (*dev->short_msg->unex)(rhandle, from_grank, pkt_ptr);
	      break;
	      
	    case MPID_PKT_REQUEST_SEND:
	      MXMPI_DEBUG_MX_RECV_PRINT("unexpected", "Request_to_send", 
				       from_grank);
	      DEBUG_TEST_FCN(dev->rndv->unex,"dev->rndv->unex");
	      err = (*dev->rndv->unex)(rhandle, from_grank, pkt_ptr);
	      break;
	      
	    default:
	      fprintf(stderr, "[%d] Internal error: msg packet discarded "
		      "(%s:%d)\n", MPID_MyWorldRank, __FILE__, __LINE__);
	    }
	}
    }
  else
    {
      switch (pkt_ptr->head.mode)
	{
	case MPID_PKT_OK_TO_SEND:
	  MXMPI_DEBUG_MX_RECV_PRINT("control", "Ok_to_send", from_grank);
	  DEBUG_TEST_FCN(dev->rndv->do_ack,"dev->rndv->do_ack");
	  err = (*dev->rndv->do_ack)(pkt_ptr, from_grank);
	  break;
	
	case MPID_PKT_DONE_SEND:
	  {
	    unsigned int chunk_size;
	    void * recvid;
	    void * done_target_ptr;
	    
	    MXMPI_DEBUG_MX_RECV_PRINT("control", "Send_done", from_grank);
	    MPID_AINT_GET(recvid, pkt_ptr->done_pkt.recv_id);
	    MPID_AINT_GET(done_target_ptr, 
			  pkt_ptr->done_pkt.done_target_ptr);
	    rhandle = (MPIR_RHANDLE *)recvid;
	    mxmpi_debug_assert(rhandle != NULL);
	    mxmpi_debug_assert(done_target_ptr != NULL);
	    chunk_size = (rhandle->mx.current_expected
			  - rhandle->mx.current_done);
	    if (chunk_size > MXMPI_MAX_PUT_LENGTH)
	      {
		chunk_size = MXMPI_MAX_PUT_LENGTH;
	      }
	    
#if MX_DISABLE_REGISTRATION
	    memcpy((void *)((unsigned long)rhandle->buf
			    + rhandle->mx.current_done),
		   done_target_ptr, chunk_size);
	    mxmpi_free_bounce_buffer((unsigned long)rhandle->buf
                                    + rhandle->mx.current_done,
                                    (unsigned long)chunk_size);
	    
#else
	    mxmpi_debug_assert((unsigned long)done_target_ptr 
			      == ((unsigned long)rhandle->buf
				  + rhandle->mx.current_done));
	    mxmpi_unuse_interval((unsigned long)rhandle->buf
				+ rhandle->mx.current_done,
				(unsigned long)chunk_size);
#endif
	    if (mxmpi.send_buf_fifo_head != NULL)
	      {
		mxmpi_flush_fifo_send();
	      }
	    
	    rhandle->mx.current_done += chunk_size;
	    rhandle->mx.in_pipe--;
	    
	    mxmpi_debug_assert(rhandle->mx.current_done 
			      <= rhandle->mx.current_expected);
	    if (rhandle->mx.current_done == rhandle->mx.current_expected)
	      {
		MXMPI_DEBUG_CHECKSUM_CHECK("Large message",
					  (void *)(rhandle->buf),
					  from_grank,
					  &(rhandle->mx.cksum));
		rhandle->is_complete = 1;
 	        if (rhandle->finish)
		  {
		    (rhandle->finish)(rhandle);
		  }
	      }
	    else
	      {
		mxmpi_debug_assert(rhandle->buf != NULL);
		mxmpi_debug_assert(rhandle->send_id != NULL);
		err = MPID_CH_Rndvn_ok_to_send(rhandle);
	      }
	  }
	  break;
	  
	case MPID_PKT_ANTI_SEND:
	  MPID_SendCancelOkPacket(pkt_ptr, from_grank); 
	  break;
	  
	case MPID_PKT_ANTI_SEND_OK:
	  MPID_RecvCancelOkPacket(pkt_ptr, from_grank); 
	  break;
	  
	default:
	  {
	    unsigned char * pktp;
	    
	    fprintf(stderr, "[%d] Packet type %d (0x%x) is unknown "
		    "%s:%d!\n",  MPID_MyWorldRank, pkt_ptr->head.mode, 
		    pkt_ptr->head.mode,__FILE__, __LINE__);
	    pktp = (unsigned char *)pkt_ptr;
	    fprintf(stderr, "[%d]: acket dump: ptr=0x%lx buf=0x%lx "
		    "0x%x%x%x%x%x%x%x%x\n", MPID_MyWorldRank, 
		    (unsigned long)pkt_ptr, (unsigned long)buf,
		    pktp[0], pktp[1], pktp[2], pktp[3],
		    pktp[4], pktp[5], pktp[6], pktp[7]);
	    
	    if ((pktp[0] == 0xaa) && (pktp[1] == 0xaa)  &&
		(pktp[2] == 0xaa) && (pktp[3] == 0xaa))
	      {
		fprintf(stderr, "[%d]: MX could not translate a virtual "
			"address\n", MPID_MyWorldRank);
		mxmpi_abort (0);
	      }
	  }
	}
    }
  
  DEBUG_PRINT_MSG("Exiting check_incoming");
  return err;
}


mx_inline int 
mxmpi_net_lookup(MPID_Device * dev, int blocking)
{
  int fast, err, total;
  void * ptr;
  mx_recv_event_t *event;
  
  err = -1;
  total = 0;
  while (1)
    {
      if (blocking == MPID_NOTBLOCKING)
	{
	  event = mx_receive(mxmpi_mx_port);
	}
      else
	{
	  event = (mxmpi.mx_receive_mode)(mxmpi_mx_port);
	}

      fast = 0;
      switch (mx_ntohc(event->recv.type))
	{
	case MX_NO_RECV_EVENT:
	  return err;
	  
	case MX_FAST_RECV_EVENT:
	case MX_FAST_HIGH_RECV_EVENT:
	case MX_FAST_PEER_RECV_EVENT:
	case MX_FAST_HIGH_PEER_RECV_EVENT:
	  fast = 1;
	  
	case MX_RECV_EVENT:
	case MX_HIGH_RECV_EVENT:
	case MX_PEER_RECV_EVENT:
	case MX_HIGH_PEER_RECV_EVENT:
	  if (fast == 1)
	    {
	      ptr = mx_ntohp(event->recv.message);
	    }
	  else
	    {
	      ptr = mx_ntohp(event->recv.buffer); 
	    }
	  
	  mxmpi_debug_assert(mx_ntohc(event->recv.size) == MXMPI_PKT_MX_SIZE);
	  mxmpi_packet_recv_event(dev, ptr);
	  mx_provide_receive_buffer(mxmpi_mx_port,
				    mx_ntohp(event->recv.buffer),
				    MXMPI_PKT_MX_SIZE, MX_LOW_PRIORITY);
          total++;
	  if ((blocking == MPID_NOTBLOCKING) && (total >= 10))
	    { 
	      err = MPI_SUCCESS;
	    }
	  else
	    {
	      return MPI_SUCCESS;
	    }
	  break;
	  
	default:
          callback_flag = 0;
	  mx_unknown(mxmpi_mx_port, event);
          if (callback_flag == 1)
            {
	     if (blocking == MPID_NOTBLOCKING)
	       { 
		 err = MPI_SUCCESS;
	       }
	     else
	       {
		 return MPI_SUCCESS;
	       }
	    }
	}
    }
}


int 
MPID_CH_Check_incoming(MPID_Device *dev, MPID_BLOCKING_TYPE blocking)
{
  int found;
 
  found = -1;
  MXMPI_PROGRESSION_LOCK();
  found = mxmpi_net_lookup(dev, blocking);
  MXMPI_PROGRESSION_UNLOCK();

  return found;
}


void 
mxmpi_init(int *argc,char ***argv)
{
  mx_status_t status;
  struct mxmpi_send_buf * mxmpi_send_buf_ptr;
  struct mxmpi_send_buf * previous_send_buf_ptr;
  void * ptr;
  unsigned int i;
  
#if MXMPI_DEBUG_OUTPUT_FILE
  char output_name[256];
  
  sprintf(output_name, "%d.out", MPID_MyWorldRank); 
  mxmpi.debug_output_filedesc = fopen(output_name, "w+");
#else
  mxmpi.debug_output_filedesc = stderr;
#endif
  
  mxmpi_debug_assert (mxmpi_mx_port != NULL);

  mxmpi.pending_put_callback = 0;
  mxmpi.pending_send_callback = 0;
  mxmpi.unexpected_short = 0;

  mxmpi_debug_assert(mxmpi.my_node_id == mxmpi.node_ids[MPID_MyWorldRank]);
  
  for (i=0; i<MPID_MyWorldSize; i++)
    {
      mxmpi.pending_sends[i] = 0;
      mxmpi.dropped_sends[i] = 0;
    }
}


void 
mxmpi_finish(MPID_Device * dev)
{
  int i;
  
  while (SOMETHING)
    {
      (*dev->check_device)(dev, MPID_NOTBLOCKING);
    }
  
  mx_close_port(mxmpi_mx_port);
  
  for (i=0; i<MPID_MyWorldSize; i++)
    {
      if (mxmpi.host_names[i] != NULL)
	{
	  free (mxmpi.host_names[i]);
	  free (mxmpi.exec_names[i]);
	}
      else
	{
	  i = MPID_MyWorldSize;
	}
    }
  
  fclose(mxmpi.debug_output_filedesc);
  fflush (stdout);
  fflush (stderr);
}


void 
mxmpi_printf(char *format, ...)
{
  va_list ap;
  
  va_start(ap, format);
  vfprintf(stderr,format, ap);
  va_end(ap);
}


void 
mxmpi_abort (int code)
{
  static int aborted = 0;
  int sockfd, count, i;
  char buffer[32];
  
  if (aborted == 0)
    {
      if (mxmpi.mpd)
	{
	  MPD_Abort (code);
	}
      else
	{
	  fflush (stderr);
	  fflush (stdout);
	  aborted = 1;
      
	  /* get a socket */
	  sockfd = socket (AF_INET, SOCK_STREAM, 0);
	  if (sockfd < 0)
	    {
	      fprintf (stderr, "[%d] Error: Unable to open a socket !\n", 
		       MPID_MyWorldRank);
	      exit (-1);
	    }
	  
	  /* connect to the master */
	  i = 10;
	  while (connect (sockfd, (struct sockaddr *) (&(mxmpi.master_addr)), 
			  sizeof (mxmpi.master_addr)) < 0)
	    {
#ifdef WIN32
	      Sleep (1);
#else
	      sleep(1); 
#endif
	      if (i == 0)
		{
		  fprintf (stderr, "[%d] Error: Unable to connect to "
			   "the master to abort !\n", MPID_MyWorldRank);
		  exit (-1);
		}
	      i--;
	    }
	  
	  /* send the abort message */
	  count = 0;
	  sprintf (buffer, "<<<ABORT_%d_ABORT>>>", mxmpi.magic);
	  while (count < strlen (buffer))
	    {
	      i = write (sockfd, &(buffer[count]), strlen (buffer) - count);
	      if (i < 0)
		{
		  fprintf (stderr, "[%d] Error: write to socket failed !\n", 
			   MPID_MyWorldRank);
		  exit (-1);;
		}
	      count += i;
	    }
	  
	  close (sockfd);
	  exit (-1);
	}
    }
}


/* Created a routine for remote PID's for TotalView debugger */
int 
mxmpi_proc_info (int i, char **host, char **exename)
{
  int pid;
  
  pid = mxmpi.mpi_pids[i];
  *host = mxmpi.host_names[i];
  *exename = mxmpi.exec_names[i];
  return (pid);
}
